import java.io.IOException;
import java.util.ArrayList;

/*
    Name: DictionaryBot

    The strategy of DictionaryBot is simply to take every hint into account.
    Since the bot knows all the words, we can guess one of the words and then use the resulting hints to rule out words that don't fit the hints.
    For example: if a particular letter is not in the word, then rule out all words with that letter.
    If a letter is somewhere in the word, but not in the correct spot. Remove all words without that letter, and all words with that letter in that position.

    By strictly following the hints, we should get to the final answer very quickly without making any outside assumptions.
    I made my bot work to find all the valid Wordle words from the actual game and I found that on average word takes around 6-7 guesses.
    Then I put it all on a graph: https://bit.ly/3GwNzqK
*/

public class DictionaryBot implements WordlePlayer {

    ArrayList<String> knownWords;
    boolean hasGuess = true;
    int guessNumber = 0;
    String nextGuess = "";

    public DictionaryBot() {}

    @Override
    public void beginGame(Wordle game) {
        knownWords = game.getKnownWords();
    }

    @Override
    public boolean hasNextGuess() {
        return true;
    }

    @Override
    public String nextGuess() {
        guessNumber++;
        if (guessNumber == 1) {
            return knownWords.get(0);
        }
        return nextGuess;
    }

    @Override
    public void tell(Hint h) {
        if (h.isWin()) {
            hasGuess = false;
            return;
        }

        for (char c : h.getNotInPuzzle().toCharArray()) {
            // We have to check correctly and incorrectly placed to make sure a letter isn't there just because it's a duplicate.
            if (h.getCorrectlyPlaced().contains(String.valueOf(c))) continue;
            if (h.getIncorrectlyPlaced().contains(String.valueOf(c))) continue;
            // Remove all words that don't contain letters not in the puzzle.
            knownWords.removeIf(s->s.contains(String.valueOf(c)));
        }

        int incorrectIndex = 0;
        // Index of the incorrect letter
        for (char c : h.getIncorrectlyPlaced().toCharArray()) {
            if (c == '-') {
                incorrectIndex++;
                continue;
            }
            // This is something IntelliJ had me do to use the value in a lambda.
            int finalIncorrectIndex = incorrectIndex;
            // Remove all words that have the letter in that certain position.
            knownWords.removeIf(s->s.charAt(finalIncorrectIndex) == c);
            incorrectIndex++;
        }

        String correctRegex = "^";
        // Using a regex to match multiple characters in multiple positions.
        for (char c : h.getCorrectlyPlaced().toCharArray()) {
            if (c == '-') {
                correctRegex += "[\\w]";
            } else {
                correctRegex += "[" + c + "]";
            }
        }
        // This is something IntelliJ had me do to use the value in a lambda.
        String finalCorrectRegex = (correctRegex + "$");
        // Remove words that don't match the correctlyPlaced hint.
        knownWords.removeIf(s->!s.matches(finalCorrectRegex));

        // Guess the next word and rule it out instantly.
        nextGuess = knownWords.get(0);
        knownWords.remove(0);
    }

    public static void main(String[] args) throws IOException {
        /*Wordle wg = new WordleGame("valid-wordle-words.txt", 5, 0, Integer.MAX_VALUE);
        ArrayList<String> words = wg.getKnownWords();

        for (String word : words) {
            WordleGame game = new WordleGame("valid-wordle-words.txt", 5, 0, Integer.MAX_VALUE);
            game.initGame(word);

            DictionaryBot bot = new DictionaryBot();
            bot.beginGame(game);
            Hint h = null;
            int guess = 0;
            int maxGuesses = 20;
            while (bot.hasNextGuess() && guess < maxGuesses) {
                guess++;
                h = game.guess(bot.nextGuess());
                //((WordleHint) h).write();
                bot.tell(h);
                if (h.isWin()) {
                    break;
                }
            }
            System.out.println("Game over: bot " + (h.isWin() ? "won" : "lost") + " with " + guess + " guesses. word: " + word);
        }*/
    }

}
